#!/usr/bin/python3

import os
import time
import inspect
import threading
import sys

DEBUG_MODE = os.getenv("DEBUG", False)

# Used as a decorator to time functions
def print_timing(func):
    if not DEBUG_MODE:
        return func
    else:
        def wrapper(*arg):
            t1 = time.time()
            res = func(*arg)
            t2 = time.time()
            print('mint-common (DEBUG): %s took %0.3f ms' % (func.__qualname__, (t2 - t1) * 1000.0), flush=True, file=sys.stderr)
            return res
        return wrapper

def check_ml():
    if not DEBUG_MODE:
        return

    if threading.current_thread().name is not None:
        tid = threading.current_thread().name
    else:
        tid = str(threading.get_ident())
    fid = inspect.stack()[1][3]
    on_ml = threading.current_thread() == threading.main_thread()
    print("%s in thread: %s" % (fid, tid), flush=True, file=sys.stderr)

def debug(*args):
    if not DEBUG_MODE:
        return
    sanitized = [str(arg) for arg in args if arg is not None]
    argstr = " ".join(sanitized)
    print("mint-common (DEBUG): %s" % argstr, file=sys.stderr, flush=True)

def warn(*args):
    sanitized = [str(arg) for arg in args if arg is not None]
    argstr = " ".join(sanitized)
    print("mint-common (WARN): %s" % argstr, file=sys.stderr, flush=True)


import sys
if sys.version_info.major < 3:
    raise "python3 required"
import os

import gi
gi.require_version("AppStreamGlib", "1.0")
gi.require_version("Gtk", "3.0")
from gi.repository import AppStreamGlib, Gtk

def capitalize(string):
    if string and len(string) > 1:
        return (string[0].upper() + string[1:])
    else:
        return (string)

class PkgInfo:
    def __init__(self, pkg_hash=None):
        # Saved stuff
        self.pkg_hash = None
        if pkg_hash:
            self.pkg_hash = pkg_hash

        self.name = None
        # some flatpak-specific things
        self.refid=""
        self.remote = ""
        self.kind = 0
        self.arch = ""
        self.branch = ""
        self.commit = ""
        self.remote_url = ""

        # Display info fetched by methods always
        self.display_name = None
        self.summary = None
        self.description = None
        self.version = None
        self.icon = {}
        self.screenshots = []
        self.homepage_url = None
        self.help_url = None

        # Runtime categories
        self.categories = []

class AptPkgInfo(PkgInfo):
    def __init__(self, pkg_hash=None, apt_pkg=None):
        super(AptPkgInfo, self).__init__(pkg_hash)

        if apt_pkg:
            self.name = apt_pkg.name

    @classmethod
    def from_json(cls, json_data:dict):
        inst = cls()
        inst.pkg_hash = json_data["pkg_hash"]
        inst.name = json_data["name"]

        return inst

    def to_json(self):
        trimmed_dict = {}

        for key in ("pkg_hash",
                    "name"):
            trimmed_dict[key] = self.__dict__[key]

        return trimmed_dict

    def get_display_name(self, apt_pkg=None):
        # fastest
        if self.display_name:
            return self.display_name

        if apt_pkg:
            self.display_name = apt_pkg.name.capitalize()

        if not self.display_name:
            self.display_name = self.name.capitalize()

        self.display_name = self.display_name.replace(":i386", "")

        return self.display_name

    def get_summary(self, apt_pkg=None):
        # fastest
        if self.summary:
            return self.summary

        if apt_pkg and apt_pkg.candidate:
            candidate = apt_pkg.candidate

            summary = ""
            if candidate.summary is not None:
                summary = candidate.summary

                self.summary = capitalize(summary)

        if self.summary == None:
            self.summary = ""

        return self.summary

    def get_description(self, apt_pkg=None):
        # fastest
        if self.description:
            return self.description

        if apt_pkg and apt_pkg.candidate:
            candidate = apt_pkg.candidate

            description = ""
            if candidate.description != None:
                description = candidate.description
                description = description.replace("<p>", "").replace("</p>", "\n")
                for tags in ["<ul>", "</ul>", "<li>", "</li>"]:
                    description = description.replace(tags, "")

                self.description = capitalize(description)

        if self.description == None:
            self.description = ""

        return self.description

    def get_icon(self, pkginfo, apt_pkg=None, size=64):
        try:
            return self.icon[size]
        except:
            pass

        # Look in app-install-data and pixmaps
        for extension in ['svg', 'png', 'xpm']:
            for suffix in ['', '-icon']:
                icon_path = "/usr/share/app-install/icons/%s%s.%s" % (pkginfo.name, suffix, extension)
                if os.path.exists(icon_path):
                    self.icon[size] = icon_path
                    return self.icon[size]

                icon_path = "/usr/share/pixmaps/%s.%s" % (pkginfo.name, extension)
                if os.path.exists(icon_path):
                    self.icon[size] = icon_path
                    return self.icon[size]

        theme = Gtk.IconTheme.get_default()

        for name in [pkginfo.name, pkginfo.name.split(":")[0], pkginfo.name.split("-")[0], pkginfo.name.split(".")[-1].lower()]:
            if theme.has_icon(name):
                self.icon[size] = name
                return self.icon[size]

        return None

    def get_screenshots(self, apt_pkg=None):
        return [] # handled in mintinstall for now

    def get_version(self, apt_pkg=None):
        if self.version:
            return self.version

        if apt_pkg:
            if apt_pkg.is_installed:
                self.version = apt_pkg.installed.version
            else:
                self.version = apt_pkg.candidate.version

        if self.version == None:
            self.version = ""

        return self.version

    def get_homepage_url(self, apt_pkg=None):
        if self.homepage_url:
            return self.homepage_url

        if apt_pkg:
            if apt_pkg.is_installed:
                self.homepage_url = apt_pkg.installed.homepage
            else:
                self.homepage_url = apt_pkg.candidate.homepage

        if self.homepage_url == None:
            self.homepage_url = ""

        return self.homepage_url

    def get_help_url(self, apt_pkg=None):
        # We can only get the homepage from apt
        return ""

class FlatpakPkgInfo(PkgInfo):
    def __init__(self, pkg_hash=None, remote=None, ref=None, remote_url=None, installed=False):
        super(FlatpakPkgInfo, self).__init__(pkg_hash)

        if not pkg_hash:
            return

        self.name = ref.get_name() # org.foo.Bar
        self.remote = remote # "flathub"
        self.remote_url = remote_url

        self.installed = installed

        self.refid = ref.format_ref() # app/org.foo.Bar/x86_64/stable
        self.kind = ref.get_kind() # Will be app for now
        self.arch = ref.get_arch()
        self.branch = ref.get_branch()
        self.commit = ref.get_commit()

    @classmethod
    def from_json(cls, json_data:dict):
        inst = cls()
        inst.pkg_hash = json_data["pkg_hash"]
        inst.name = json_data["name"]
        inst.refid = json_data["refid"]
        inst.remote = json_data["remote"]
        inst.kind = json_data["kind"]
        inst.arch = json_data["arch"]
        inst.branch = json_data["branch"]
        inst.commit = json_data["commit"]
        inst.remote_url = json_data["remote_url"]

        return inst

    def to_json(self):
        trimmed_dict = {}

        for key in ("pkg_hash",
                    "name",
                    "refid",
                    "remote",
                    "kind",
                    "arch",
                    "branch",
                    "commit",
                    "remote_url"):
            trimmed_dict[key] = self.__dict__[key]

        return trimmed_dict

    def get_display_name(self, as_component=None):
        # fastest
        if self.display_name:
            return self.display_name

        if as_component:
            display_name = as_component.get_name()

            if display_name != None:
                self.display_name = capitalize(display_name)

        if self.display_name == None:
            self.display_name = self.name

        return self.display_name

    def get_summary(self, as_component=None):
        # fastest
        if self.summary:
            return self.summary

        if as_component:
            summary = as_component.get_comment()

            if summary != None:
                self.summary = summary

        if self.summary == None:
            self.summary = ""

        return self.summary

    def get_description(self, as_component=None):
        # fastest
        if self.description:
            return self.description

        if as_component:
            description = as_component.get_description()

            if description != None:
                description = description.replace("<p>", "").replace("</p>", "\n")
                for tags in ["<ul>", "</ul>", "<li>", "</li>"]:
                    description = description.replace(tags, "")
                self.description = capitalize(description)

        if self.description == None:
            self.description = ""

        return self.description

    def get_icon(self, pkginfo, as_component=None, size=64):
        try:
            return self.icon[size]
        except:
            pass

        if as_component:
            icons = as_component.get_icons()

            if icons:
                icon_to_use = None
                remote_icon = None
                first_local_icon = None
                good_size_icon = None

                for icon in icons:
                    if icon.get_kind() == AppStreamGlib.IconKind.REMOTE:
                        remote_icon = icon
                        continue

                    if icon.get_kind() in (AppStreamGlib.IconKind.LOCAL,  \
                                           AppStreamGlib.IconKind.CACHED, \
                                           AppStreamGlib.IconKind.STOCK):
                        first_local_icon = icon
                        if icon.get_height() == size or ("%dx%d" % (size, size)) in icon.get_prefix():
                            good_size_icon = icon
                            break

                icon_to_use = good_size_icon or first_local_icon or remote_icon
                kind = icon_to_use.get_kind()

                if kind in (AppStreamGlib.IconKind.LOCAL, AppStreamGlib.IconKind.CACHED):
                    self.icon[size] = os.path.join(icon_to_use.get_prefix(), icon_to_use.get_name())
                elif kind == AppStreamGlib.IconKind.REMOTE:
                    self.icon[size] = icon_to_use.get_url()
                elif kind == AppStreamGlib.IconKind.STOCK:
                    self.icon[size] = icon_to_use.get_name()

        try:
            return self.icon[size]
        except:
            return None

    def get_screenshots(self, as_component=None):
        if len(self.screenshots) > 0:
            return self.screenshots

        if as_component:
            self.screenshots = as_component.get_screenshots()

        return self.screenshots

    def get_version(self, as_component=None):
        if self.version:
            # as_component.get_release_default().get_version()
            return self.version

        if as_component:
            releases = as_component.get_releases()

            if len(releases) > 0:
                releases.sort(key=lambda r: r.get_timestamp(), reverse=True)
                version = releases[0].get_version()

                if version:
                    self.version = version

        if self.version == None:
            self.version = ""

        return self.version

    def get_homepage_url(self, as_component=None):
        if self.homepage_url:
            return self.homepage_url

        if as_component:
            url = as_component.get_url_item(AppStreamGlib.UrlKind.HOMEPAGE)

            if url != None:
                self.homepage_url = url

        return self.homepage_url

    def get_help_url(self, as_component=None):
        if self.help_url:
            return self.help_url

        if as_component:
            url = as_component.get_url_item(AppStreamGlib.UrlKind.HELP)

            if url != None:
                self.help_url = url

        if self.help_url == None:
            self.help_url = ""

        return self.help_url




#!/usr/bin/python3

import os
import sys

if __name__ == "__main__":

    # Exit if the given path does not exist
    if len(sys.argv) < 2 or not os.path.exists(sys.argv[1]):
        print("No argument or file not found")
        sys.exit(1)

    path = sys.argv[1]
    dpath = f"{path}.desktop"
    if not os.path.exists(path):
        if os.path.exists(dpath):
            path = dpath
        else:
            print("Path not found", path)
            sys.exit(1)

    os.remove(path)



import time
import threading
import datetime
import math
from pathlib import Path
import subprocess
import requests
import tempfile
import os

import gi
gi.require_version('AppStreamGlib', '1.0')
gi.require_version('Gtk', '3.0')
from gi.repository import AppStreamGlib, GLib, GObject, Gtk, Gio, Gdk

try:
    gi.require_version('Flatpak', '1.0')
    from gi.repository import Flatpak
except:
    pass

from .pkgInfo import FlatpakPkgInfo
from . import dialogs
from .dialogs import ChangesConfirmDialog, FlatpakProgressWindow
from .misc import debug, warn

class FlatpakRemoteInfo():
    def __init__(self, remote=None):
        if remote:
            self.name = remote.get_name()
            self.title = remote.get_title()
            self.summary = remote.get_comment()
            self.url = remote.get_url()
            self.disabled = remote.get_disabled()
            self.noenumerate = remote.get_noenumerate()
            if not self.title or self.title == "":
                self.title = " ".join( [word for word in self.name.split("-")])

            self.title = self.title.title()
        else:
            self.name = None
            self.title = None
            self.summary = None
            self.url = None
            self.disabled = False
            self.noenumerate = False

    @classmethod
    def from_json(cls, json_data:dict):
        inst = cls()

        inst.name = json_data["name"]
        inst.title = json_data["title"]
        inst.url = json_data["url"]
        inst.disabled = json_data["disabled"]
        inst.noenumerate = json_data["noenumerate"]

        return inst

    def to_json(self):
        return self.__dict__

_fp_sys = None

_as_pool_lock = threading.Lock()
_as_pools = {} # keyed to remote name

def get_fp_sys():
    global _fp_sys

    if _fp_sys == None:
        _fp_sys = Flatpak.Installation.new_system(None)

    return _fp_sys

ALIASES = {
}

def make_pkg_hash(ref):
    if not isinstance(ref, Flatpak.Ref):
        raise TypeError("flatpak.make_pkg_hash() must receive FlatpakRef, not %s" % type(ref))

    try:
        return "fp:%s:%s" % (ref.get_origin(), ref.format_ref())
    except Exception:
        return "fp:%s:%s" % (ref.get_remote_name(), ref.format_ref())

def _get_remote_name_by_url(fp_sys, url):
    name = None

    try:
        remotes = fp_sys.list_remotes()
    except GLib.Error:
        remotes = []

    for remote in remotes:
        remote_url = remote.get_url()
        if remote_url.endswith('/'): #flatpakrefs are often missing the trailing forward slash in the url
            remote_url = remote_url[:-1]

        if remote_url == url:
            name = remote.get_name()

    return name

def _process_remote(cache, fp_sys, remote, arch):
    remote_name = remote.get_name()

    if remote.get_disabled():
        debug("Installer: flatpak - remote '%s' is disabled, skipping" % remote_name)
        return

    debug("Installer: flatpak - updating appstream data for remote '%s'..." % remote_name)

    try:
        success = fp_sys.update_appstream_sync(remote_name, arch, None)
    except GLib.Error:
        # Not fatal..
        pass

    # get_noenumerate indicates whether a remote should be used to list applications.
    # Instead, they're intended for single downloads (via .flatpakref files)
    if remote.get_noenumerate():
        debug("Installer: flatpak - remote '%s' is marked as no-enumerate skipping package listing" % remote_name)
        return

    remote_url = remote.get_url()

    try:
        for ref in fp_sys.list_remote_refs_sync(remote_name, None):
            if ref.get_kind() == Flatpak.RefKind.RUNTIME:
                continue

            if ref.get_name().endswith("BaseApp"):
                continue

            if ref.get_name().endswith("BaseExtension"):
                continue

            if ref.get_arch() != arch:
                continue

            if ref.get_eol() is not None:
                continue

            _add_package_to_cache(cache, ref, remote_url, False)
    except GLib.Error as e:
        warn("Process remote:", e.message)

def _add_package_to_cache(cache, ref, remote_url, installed):
    pkg_hash = make_pkg_hash(ref)

    try:
        remote_name = ref.get_remote_name()
    except Exception:
        remote_name = ref.get_origin()

    try:
        pkginfo = cache[pkg_hash]

        if installed:
            pkginfo.installed = installed
    except KeyError:
        pkginfo = FlatpakPkgInfo(pkg_hash, remote_name, ref, remote_url, installed)
        cache[pkg_hash] = pkginfo

    return pkginfo

def process_full_flatpak_installation(cache):
    fp_time = time.time()

    arch = Flatpak.get_default_arch()
    fp_sys = get_fp_sys()

    flatpak_remote_infos = {}

    try:
        for remote in fp_sys.list_remotes():
            _process_remote(cache, fp_sys, remote, arch)

            remote_name = remote.get_name()

            try:
                for ref in fp_sys.list_installed_refs(None):
                    # All remotes will see installed refs, but the installed refs will always
                    # report their correct origin, so only add installed refs when they match the remote.
                    if ref.get_origin() == remote_name:
                        _add_package_to_cache(cache, ref, remote.get_url(), True)
            except GLib.Error as e:
                warn("adding packages:", e.message)

            flatpak_remote_infos[remote_name] = FlatpakRemoteInfo(remote)

    except GLib.Error as e:
        warn("Installer: flatpak - could not get remote list", e.message)
        cache = {}

    debug('Installer: Processing Flatpaks for cache took %0.3f ms' % ((time.time() - fp_time) * 1000.0))

    return cache, flatpak_remote_infos

def _load_appstream_pool(remote):
    pool = AppStreamGlib.Store()

    try:
        path = remote.get_appstream_dir().get_path()

        with open(os.path.join(path, "appstream.xml")) as f:
            pool.from_xml(f.read(), path)
    except Exception as e:
        warn("Could not load appstream info for remote '%s': %s" % (remote, str(e)))
        return

    _as_pools[remote.get_name()] = pool

def initialize_appstream():
    thread = threading.Thread(target=_initialize_appstream_thread)
    thread.start()

def _initialize_appstream_thread():
    fp_sys = get_fp_sys()

    global _as_pools
    global _as_pool_lock

    with _as_pool_lock:
        _as_pools = {}

        try:
            for remote in fp_sys.list_remotes():
                _load_appstream_pool(remote)
        except (GLib.Error, Exception) as e:
            try:
                msg = e.message
            except:
                msg = str(e)
            warn("Installer: Could not initialize appstream components for flatpaks: %s" % msg)

def get_remote_or_installed_ref(ref, remote_name):
    fp_sys = get_fp_sys()

    try:
        iref = fp_sys.get_installed_ref(ref.get_kind(),
                                        ref.get_name(),
                                        ref.get_arch(),
                                        ref.get_branch(),
                                        None)

        if iref:
            return iref
    except GLib.Error as e:
        if e.code != Flatpak.Error.NOT_INSTALLED:
            warn("Installer: Couldn't look up InstalledRef: %s" % e.message)

    try:
        rref = fp_sys.fetch_remote_ref_sync(remote_name,
                                            ref.get_kind(),
                                            ref.get_name(),
                                            ref.get_arch(),
                                            ref.get_branch(),
                                            None)
        if rref:
            return rref
    except GLib.Error as e:
        if e.code != Flatpak.Error.ALREADY_INSTALLED:
            warn("Installer: Couldn't look up RemoteRef (%s): %s" % (remote_name, e.message))

    return None

def create_pkginfo_from_as_component(comp, remote_name, remote_url):
    name = comp.get_pkgname_default()
    branch = comp.get_branch()

    bundle = comp.get_bundle_default()

    shallow_ref = Flatpak.Ref.parse(bundle.get_id())

    ref = get_remote_or_installed_ref(shallow_ref, remote_name)
    if ref is None:
        return None

    pkg_hash = make_pkg_hash(ref)
    pkginfo = FlatpakPkgInfo(pkg_hash, remote_name, ref, remote_url)
    pkginfo.installed = isinstance(ref, Flatpak.InstalledRef)

    return pkginfo

def search_for_pkginfo_as_component(pkginfo):
    asapps = _search_as_pool_by_name(pkginfo.name, pkginfo.remote)

    if asapps is None:
        return None

    for app in asapps:
        bundle = app.get_bundle_default()
        if pkginfo.refid == bundle.get_id():
            return app

    return None

def _search_as_pool_by_name(name, remote):
    comps = []

    with _as_pool_lock:
        try:
            pool = _as_pools[remote]
        except Exception as e:
            return None

        comps = pool.get_apps_by_id(name)

        if comps == []:
            comps = pool.get_apps_by_id(name + ".desktop")

    if len(comps) > 0:
        return comps
    else:
        return None

def _get_system_theme_matches():
    fp_sys = get_fp_sys()
    arch = Flatpak.get_default_arch()

    theme_refs = []

    gtksettings = Gtk.Settings.get_default()

    icon_theme = "org.freedesktop.Platform.Icontheme.%s" % gtksettings.props.gtk_icon_theme_name
    gtk_theme = "org.gtk.Gtk3theme.%s" % gtksettings.props.gtk_theme_name

    def sortref(ref):
        try:
            val = float(ref.get_branch())
        except ValueError:
            val = 9.9

        return val

    for name in (icon_theme, gtk_theme):
        theme_ref = None

        for remote in fp_sys.list_remotes():
            if remote.get_nodeps():
                continue
            remote_name = remote.get_name()

            try:
                debug("Looking for theme %s in %s" % (name, remote_name))

                all_refs = fp_sys.list_remote_refs_sync(remote_name, None)
                matching_refs = []

                for listed_ref in all_refs:
                    if listed_ref.get_name() == name:
                        matching_refs.append(listed_ref)

                if not matching_refs:
                    continue

                # Sort highest version first.
                matching_refs = sorted(matching_refs, key=sortref, reverse=True)

                for matching_ref in matching_refs:
                    if matching_ref.get_arch() != arch:
                        continue

                    theme_ref = matching_ref
                    debug("Found theme ref '%s' in remote %s" % (theme_ref.format_ref(), remote_name))
                    break

            except GLib.Error as e:
                theme_ref = None
                warn("Error finding themes for flatpak: %s" % e.message)

        if theme_ref:
            theme_refs.append(theme_ref)

    return theme_refs

def _get_related_refs_for_removal(parent_pkginfo):
    return_refs = []

    # .Locale files
    related_refs = get_fp_sys().list_installed_related_refs_sync(parent_pkginfo.remote,
                                                                 parent_pkginfo.refid,
                                                                 None)
    return related_refs

def select_packages(task):
    task.transaction = FlatpakTransaction(task)

    debug("Installer: Calculating changes required for Flatpak package: %s" % task.pkginfo.name)

def select_updates(task):
    task.transaction = FlatpakTransaction(task)

    debug("Installer: Calculating Flatpak updates.")

class FlatpakTransaction():
    def __init__(self, task):
        self.task = task
        self.transaction = Flatpak.Transaction.new_for_installation(get_fp_sys(), task.cancellable)
        self.item_count = 0
        self.current_count = 0

        self.transaction_ready = False
        self.current_fp_progress = None
        self.op_error = None

        self.start_transaction = threading.Event()

        self.transaction.connect("ready", self.on_transaction_ready)
        self.transaction.connect("new-operation", self._new_operation)
        self.transaction.connect("operation-done", self._operation_done)
        self.transaction.connect("operation-error", self._operation_error)
        self.transaction.connect("add-new-remote", self._transaction_add_new_remote)
        self.transaction.connect("end-of-lifed-with-rebase", self._ref_eoled_with_rebase)

        # Runtimes explicitly installed are 'pinned' - which means they'll never be automatically
        # removed due to being unused. Addons are useless without the apps they're for, so we can
        # disable pinning for them.
        if self.task.is_addon_task:
            self.transaction.set_disable_auto_pin(True)

        thread = threading.Thread(target=self._transaction_thread, name="flatpak-transaction-thread")
        thread.start()

    def _transaction_thread(self):
        try:
            if self.task.type == "install":
                self.transaction.add_install(self.task.pkginfo.remote,
                                             self.task.pkginfo.refid,
                                             None)
            elif self.task.type == "remove":
                self.transaction.add_uninstall(self.task.pkginfo.refid)

                if not self.task.is_addon_task:
                    for related_ref in _get_related_refs_for_removal(self.task.pkginfo):
                        self.transaction.add_uninstall(related_ref.format_ref())
            else:
                try:
                    all_updates = get_fp_sys().list_installed_refs_for_update(self.task.cancellable)

                    if self.task.initial_refs_to_update != []:
                        for ref in self.task.initial_refs_to_update:
                            # Sometimes it turns out we have a new package to install that is not part of
                            # another package's pulled-in dependencies. It ends up as a selectable update in
                            # mintupdate. Once it does, though, we have to find its associated package in
                            # the original update list so we know which remote to try and pull it from.
                            #
                            # FIXME: this is because select_updates only takes a ref string. It could take a a remote
                            # or installed- ref instead.
                            if not ref_is_installed(Flatpak.Ref.parse(ref)):
                                for installed_ref in all_updates:
                                    related_refs = get_fp_sys().list_remote_related_refs_sync(installed_ref.get_origin(),
                                                                                              installed_ref.format_ref(),
                                                                                              self.task.cancellable)
                                    for related_ref in related_refs:
                                        if related_ref.format_ref() == ref:
                                            self.transaction.add_install(installed_ref.get_origin(), ref, None)
                            else:
                                self.transaction.add_update(ref, None, None)

                    else:
                        for ref in all_updates:
                            self.transaction.add_update(ref.format_ref(), None, None)
                except GLib.Error as e:
                    warn("Problem checking installed flatpaks updates: %s" % e.message)
                    raise


            # Always install the corresponding theme if we didn't already
            # have it.
            if self.task.type != "remove":
                if self.task.asapp is not None and self.task.asapp.get_kind() != AppStreamGlib.AppKind.ADDON:
                    for theme_ref in _get_system_theme_matches():
                        try:
                            self.transaction.add_install(theme_ref.get_remote_name(),
                                                         theme_ref.format_ref(),
                                                         None)
                        except GLib.Error as e:
                            if e.code == Flatpak.Error.ALREADY_INSTALLED:
                                continue
                            else:
                                raise

            # Simulate the install, cancel once ops are generated.

        except GLib.Error as e:
            self.on_transaction_error(e)

        try:
            self.transaction.run(self.task.cancellable)
        except GLib.Error as e:
            self.on_transaction_error(e)

        self.on_transaction_finished()

    def save_ref_current_version(self, ref):
        version = "<unknown>"

        try:
            version = ref.get_appdata_version()

            if version is None:
                version = ref.get_latest_commit()
        except:
            # not installed
            version = "installing"
            pass

        debug("Adding prior ref version for %s: %s" % (ref.format_ref(), version))
        self.task.ref_prior_versions_dict[ref.format_ref()] = version

    def on_transaction_error(self, error):
        if not self.op_error:
            if error.code in (Flatpak.Error.ABORTED, Gio.IOErrorEnum.CANCELLED):
                return

        if self.task.info_ready_status == self.task.STATUS_NONE:
            self.task.info_ready_status = self.task.STATUS_UNKNOWN

        if not self.transaction_ready:
            self.task.handle_error(error, info_stage=True)
        else:
            if self.op_error:
                self.task.handle_error(self.op_error)
            else:
                self.task.handle_error(error)

    def on_transaction_finished(self):
        get_fp_sys().drop_caches(None)

        # If an op failed, show an error, even though we 'finished successfully'
        if self.task.type == self.task.UPDATE_TASK and self.op_error:
            self.on_transaction_error(self.op_error)

        if self.task.error_message:
            self.task.call_error_cleanup_callback()
        else:
            self.task.call_finished_cleanup_callback()

    def on_transaction_progress(self, progress):
        package_chunk_size = 1.0 / self.item_count
        partial_chunk = (progress.get_progress() / 100.0) * package_chunk_size
        actual_progress = math.floor(((self.current_count * package_chunk_size) + partial_chunk) * 100.0)
        if self.task.client_progress_cb:
            Gdk.threads_add_idle(GLib.PRIORITY_DEFAULT,
                                 self.task.client_progress_cb,
                                 self.task.pkginfo,
                                 actual_progress,
                                 progress.get_is_estimating(),
                                 progress.get_status())

    def _new_operation(self, transaction, op, progress):
        progress.set_update_frequency(500)
        progress.connect("changed", self.on_transaction_progress)

    def _operation_error(self, transaction, operation, error, details):
        # Set error from the failing operation - Overall transaction errors from real failure
        # use the same ABORTED code. The op error will be more specific and useful (and let us
        # distinguish cancel from fail).

        # If the user cancelled the operation, cancel the transaction, but don't log it.
        if error.code == Gio.IOErrorEnum.CANCELLED:
            return False

        self.op_error = error
        self.log_operation_result(operation, None, error)

        # Don't abort remaining operations if we're doing updates.
        return self.task.type == self.task.UPDATE_TASK

    def _operation_done(self, transaction, operation, commit, result, data=None):
        self.log_operation_result(operation, result)

        self.current_count += 1
        if self.current_count < self.item_count:
            return

    def on_transaction_ready(self, transaction):
        self.transaction_ready = True

        try:
            fp_sys = get_fp_sys()

            dl_size = 0
            disk_size = 0

            for op in self.transaction.get_operations():
                ref = Flatpak.Ref.parse(op.get_ref())
                op_type = op.get_operation_type()

                if op_type == Flatpak.TransactionOperationType.INSTALL:
                    dl_size += op.get_download_size()
                    disk_size += op.get_installed_size()

                    self.save_ref_current_version(ref)
                    self._add_to_list(self.task.to_install, ref)
                elif op_type == Flatpak.TransactionOperationType.UNINSTALL:
                    iref = fp_sys.get_installed_ref(ref.get_kind(),
                                                    ref.get_name(),
                                                    ref.get_arch(),
                                                    ref.get_branch(),
                                                    None)
                    disk_size -= iref.get_installed_size()

                    self.save_ref_current_version(iref)
                    self._add_to_list(self.task.to_remove, ref)
                else: # update
                    iref = fp_sys.get_installed_ref(ref.get_kind(),
                                                    ref.get_name(),
                                                    ref.get_arch(),
                                                    ref.get_branch(),
                                                    None)

                    current_installed_size = iref.get_installed_size()
                    new_installed_size = op.get_installed_size()
                    dl_size += op.get_download_size()
                    disk_size += new_installed_size - current_installed_size

                    self.save_ref_current_version(iref)
                    self._add_to_list(self.task.to_update, ref)

            self.task.download_size = dl_size
            if disk_size > 0:
                self.task.install_size = disk_size
            else:
                self.task.freed_size = abs(disk_size)

        except Exception as e:
            # Something went wrong, bail out
            self.task.info_ready_status = self.task.STATUS_BROKEN
            self.task.handle_error(e, info_stage=True)
            return False # Close 'ready' callback, cancel.

        if len(self.task.to_install) > 0:
            debug("For install:")
            for ref in self.task.to_install:
                debug(ref.format_ref())
        if len(self.task.to_remove) > 0:
            debug("For removal:")
            for ref in self.task.to_remove:
                debug(ref.format_ref())
        if len(self.task.to_update) > 0:
            debug("For updating:")
            for ref in self.task.to_update:
                debug(ref.format_ref())

        self.item_count = len(self.task.to_install + self.task.to_remove + self.task.to_update)

        self.task.info_ready_status = self.task.STATUS_OK
        self.task.confirm = self._confirm_transaction
        self.task.cancel = self._cancel_transaction
        self.task.execute = self._execute_transaction
        self.task.call_info_ready_callback()

        self.start_transaction.wait()

        if self.task.cancellable.is_cancelled():
            return False

        return True

    def _transaction_add_new_remote(self, transaction, reason_code, from_id, suggested_remote_name, url, data=None):
        if reason_code == Flatpak.TransactionRemoteReason.GENERIC_REPO:
            reason = "The remote has additional apps."
        elif reason_code == Flatpak.TransactionRemoteReason.RUNTIME_DEPS:
            reason = "The remote has runtimes needed for the application."
        else:
            reason = "Reason unknown"

        debug("Adding new remote '%s' (%s) for %s: %s" % (suggested_remote_name, url, from_id, reason))
        return True

    def _ref_eoled_with_rebase(self, transaction, remote, ref, reason, rebased_to_ref, prev_ids):
        warn("%s is EOL (%s). Replacing with %s" % (ref, reason, rebased_to_ref))

        try:
            transaction.add_uninstall(ref)
        except:
            pass
        try:
            transaction.add_rebase(rebased_to_ref)
        except:
            debug("No new ref to rebase to, using the eol'd one")
            return False

        return True

    def _add_to_list(self, ref_list, ref):
        ref_str = ref.format_ref()

        for existing_ref in ref_list:
            if ref_str == existing_ref.format_ref():
                debug("Skipping %s, already added to task" % ref_str)
                return

        ref_list.append(ref)

    def _get_runtime_ref_from_remote_metadata(self, remote_name, ref_str):
        runtime_ref = None

        ref = Flatpak.Ref.parse(ref_str)

        meta = get_fp_sys().fetch_remote_metadata_sync(remote_name, ref, None)
        data = meta.get_data().decode()

        keyfile = GLib.KeyFile.new()
        keyfile.load_from_data(data, len(data), GLib.KeyFileFlags.NONE)

        runtime = keyfile.get_string("Application", "runtime")
        runtime_ref = Flatpak.Ref.parse("runtime/%s" % runtime)

        return runtime_ref.format_ref()

    def _confirm_transaction(self):
        # only show a confirmation if:
        # - (install/remove) Additional changes are triggered for more than just the selected package.
        # - we're updating all available packages
        # - the packages specifically selected to be updated (initial_refs_to_update) trigger additional package installs/updates/removals
        total_count = len(self.task.to_install + self.task.to_remove + self.task.to_update)
        additional = False

        if total_count == 0:
            debug("No work to perform now - are you online still?")
            # FIXME: If the network's down, flatpak doesn't consider not being able to access remote refs as fatal, since it's an update
            # and they're already installed. We should popup a message to say so.
            return False

        if self.task.type in (self.task.INSTALL_TASK, self.task.UNINSTALL_TASK) and total_count > 1:
            additional = True
        elif self.task.type == self.task.UPDATE_TASK:
            if len(self.task.initial_refs_to_update) == 0 or (total_count - len(self.task.initial_refs_to_update)) > 0:
                additional = True

        if additional:
            dia = ChangesConfirmDialog(None, self.task, parent=self.task.parent_window)
            res = dia.run()
            dia.hide()
            dia.destroy()
            return res == Gtk.ResponseType.OK
        else:
            return True

    def _cancel_transaction(self):
        self.task.cancellable.cancel()
        self.start_transaction.set()

    def _execute_transaction(self):
        if self.task.cancellable.is_cancelled():
            return False

        if self.task.client_progress_cb != None:
            self.task.has_window = True
            GLib.idle_add(self.task.client_progress_cb, self.task.pkginfo, 0, True, " : ")
        else:
            GLib.idle_add(self._show_progress_window, self.task)

        self.start_transaction.set()

    def _show_progress_window(self, task):
        progress_window = FlatpakProgressWindow(self.task)
        progress_window.present()

    def get_operations(self):
        return self.transaction.get_operations()

    def log_operation_result(self, operation, result, error=None):
        log_timestamp = datetime.datetime.now().strftime("%F::%T")
        basic_ref = Flatpak.Ref.parse(operation.get_ref())

        old_version = self.task.ref_prior_versions_dict[basic_ref.format_ref()]

        new_version = "<none>"
        if operation.get_operation_type() in (Flatpak.TransactionOperationType.INSTALL, Flatpak.TransactionOperationType.UPDATE):
            try:
                iref = get_fp_sys().get_installed_ref(basic_ref.get_kind(),
                                                      basic_ref.get_name(),
                                                      basic_ref.get_arch(),
                                                      basic_ref.get_branch(),
                                                      None)
                new_version = iref.get_appdata_version()

                if new_version is None:
                    new_version = iref.get_latest_commit()
            except Exception as e:
                pass
        else:
            new_version = "removed"

        if error is None:
            log_entry = "%s::%s::%s::%s::%s::%s" % (log_timestamp,
                                               basic_ref.get_kind().value_nick,
                                               Flatpak.transaction_operation_type_to_string(operation.get_operation_type()),
                                               basic_ref.get_name(),
                                               old_version,
                                               new_version)
        else:
            log_entry = "%s::%s::%s::%s::%s::FAILED: (%d): %s" % (log_timestamp,
                                               basic_ref.get_kind().value_nick,
                                               Flatpak.transaction_operation_type_to_string(operation.get_operation_type()),
                                               basic_ref.get_name(),
                                               old_version,
                                               error.code,
                                               error.message)

        debug("Logging: %s" % log_entry)
        self.task.transaction_log.append(log_entry)

def list_updated_pkginfos(cache):
    fp_sys = get_fp_sys()

    updated = []

    try:
        updates = fp_sys.list_installed_refs_for_update(None)
    except GLib.Error as e:
        warn("Installer: flatpak - could not get updated flatpak refs")
        return []

    for ref in updates:
        pkg_hash = make_pkg_hash(ref)

        try:
            updated.append(cache[pkg_hash])
        except KeyError:
            pass

    return updated

def get_updated_theme_refs():
    fp_sys = get_fp_sys()

    if not fp_sys.list_installed_refs_by_kind(Flatpak.RefKind.APP, None):
        return []

    return _get_system_theme_matches()

def find_pkginfo(cache, string, remote=None):
    for key in cache.get_subset_of_type("f").keys():
        candidate = cache[key]
        if string == candidate.name:
            if remote is None:
                return candidate
            elif candidate.remote == remote:
                return candidate

    return None

def generate_uncached_pkginfos(cache):
    fp_sys = get_fp_sys()

    try:
        for remote in fp_sys.list_remotes():
            remote_name = remote.get_name()

            for ref in fp_sys.list_installed_refs(None):
                # All remotes will see installed refs, but the installed refs will always
                # report their correct origin, so only add installed refs when they match the remote.
                if ref.get_origin() == remote_name:
                    _add_package_to_cache(cache, ref, remote.get_url(), True)

    except GLib.Error as e:
        warn("Installer: flatpak - could not check for uncached pkginfos", e.message)

def _ref_is_installed(kind, name, arch, branch):
    fp_sys = get_fp_sys()

    try:
        iref = fp_sys.get_installed_ref(kind,
                                        name,
                                        arch,
                                        branch,
                                        None)

        if iref:
            return True
    except GLib.Error:
        pass

    return False

def ref_is_installed(ref):
    return _ref_is_installed(ref.get_kind(),
                             ref.get_name(),
                             ref.get_arch(),
                             ref.get_branch())

def pkginfo_is_installed(pkginfo):
    return _ref_is_installed(pkginfo.kind,
                             pkginfo.name,
                             pkginfo.arch,
                             pkginfo.branch)

def list_remotes():
    fp_sys = get_fp_sys()

    remotes = []

    try:
        for remote in fp_sys.list_remotes():
            remotes.append(FlatpakRemoteInfo(remote))

    except GLib.Error as e:
        warn("Installer: flatpak - could not fetch remote list", e.message)
        remotes = []

    return remotes

def get_pkginfo_from_file(cache, file, callback):
    thread = threading.Thread(target=_pkginfo_from_file_thread, args=(cache, file, callback))
    thread.start()

def _pkginfo_from_file_thread(cache, file, callback):
    fp_sys = get_fp_sys()

    path = file.get_path()

    if path == None:
        warn("Installer: flatpak - no valid .flatpakref path provided")
        return None

    ref = None
    pkginfo = None
    remote_name = None

    with open(path) as f:
        contents = f.read()

        b = contents.encode("utf-8")
        gb = GLib.Bytes(b)

        new_remote = False

        try:
            kf = GLib.KeyFile()
            if kf.load_from_file(path, GLib.KeyFileFlags.NONE):
                name = kf.get_string("Flatpak Ref", "Name")
                url = kf.get_string("Flatpak Ref", "Url")

                try:
                    branch = kf.get_string("Flatpak Ref", "Branch")
                except GLib.Error as e:
                    if e.code == GLib.KeyFileError.KEY_NOT_FOUND:
                        warn("Installer: flatpak - flatpakref file doesn't have a Branch key, maybe nightly or testing.")
                        branch = None

                remote_name = _get_remote_name_by_url(fp_sys, url)

                if name and remote_name:
                    ref = Flatpak.RemoteRef(remote_name=remote_name,
                                            kind=Flatpak.RefKind.APP,
                                            arch=Flatpak.get_default_arch(),
                                            branch=branch,
                                            name=name)
                    warn("Installer: flatpak - using existing remote '%s' for flatpakref file install" % remote_name)
                else: #If Flatpakref is not installed already
                    try:
                        warn("Installer: flatpak - trying to install new remote for flatpakref file")
                        ref = fp_sys.install_ref_file(gb, None)
                        fp_sys.drop_caches(None)

                        remote_name = ref.get_remote_name()
                        new_remote = True
                        warn("Installer: flatpak - added remote '%s'" % remote_name)
                    except GLib.Error as e:
                        if e.code != Gio.DBusError.ACCESS_DENIED: # user cancelling auth prompt for adding a remote
                            warn("Installer: could not add new remote to system: %s" % e.message)
                            dialogs.show_flatpak_error(e.message)
        except GLib.Error as e:
            warn("Installer: flatpak - could not parse flatpakref file: %s" % e.message)
            dialogs.show_flatpak_error(e.message)

        if ref:
            try:
                remote = fp_sys.get_remote_by_name(remote_name, None)

                # We only process if it's not a new remote, otherwise our appstream data
                # will be out of sync with our package cache until we refresh the cache. This
                # can affect versioning especially.
                if new_remote:
                    _process_remote(cache, fp_sys, remote, Flatpak.get_default_arch())

                # Add the ref to the cache, so we can work with it like any other in mintinstall
                pkginfo = _add_package_to_cache(cache, ref, remote.get_url(), False)

                # Fetch the appstream info for the ref
                global _as_pools

                with _as_pool_lock:
                    if remote_name not in _as_pools.keys():
                        _load_appstream_pool(_as_pools, remote)

                # Some flatpakref files will have a pointer to a runtime .flatpakrepo file
                # We need to process and possibly add that remote as well.

                kf = GLib.KeyFile()
                if kf.load_from_file(path, GLib.KeyFileFlags.NONE):
                    try:
                        url = kf.get_string("Flatpak Ref", "RuntimeRepo")
                    except GLib.Error:
                        url = None

                    if url:
                        # Fetch the .flatpakrepo file
                        r = requests.get(url, stream=True)

                        file = tempfile.NamedTemporaryFile(delete=False)

                        with file as fd:
                            for chunk in r.iter_content(chunk_size=128):
                                fd.write(chunk)

                        # Get the true runtime url from the repo file
                        runtime_repo_url = _get_repofile_repo_url(file.name)

                        if runtime_repo_url:
                            existing = False

                            path = Path(file.name)
                            runtime_remote_name = Path(url).stem

                            # Check if the remote is already installed
                            for remote in fp_sys.list_remotes(None):
                                # See comments below in _remote_from_repo_file_thread about get_noenumerate() use.
                                if remote.get_url() == runtime_repo_url and not remote.get_noenumerate():
                                    warn("Installer: flatpak - runtime remote '%s' already in system, skipping" % runtime_remote_name)
                                    existing = True
                                    break

                            if not existing:
                                warn("Installer: Adding additional runtime remote named '%s' at '%s'" % (runtime_remote_name, runtime_repo_url))

                                cmd_v = ['flatpak',
                                         'remote-add',
                                         '--from',
                                         runtime_remote_name,
                                         file.name]

                                add_repo_proc = subprocess.Popen(cmd_v)
                                retcode = add_repo_proc.wait()

                                fp_sys.drop_caches(None)
                        os.unlink(file.name)
            except GLib.Error as e:
                warn("Installer: could not process .flatpakref file: %s" % e.message)
                dialogs.show_flatpak_error(e.message)

    GLib.idle_add(callback, pkginfo, priority=GLib.PRIORITY_DEFAULT)

def add_remote_from_repo_file(cache, file, callback):
    thread = threading.Thread(target=_remote_from_repo_file_thread, args=(cache, file, callback))
    thread.start()

def _remote_from_repo_file_thread(cache, file, callback):
    try:
        path = Path(file.get_path())
    except TypeError:
        warn("Installer: flatpak - no valid .flatpakrepo path provided")
        return

    fp_sys = get_fp_sys()

    # Make sure the remote isn't already setup (even under a different name)
    # We need to exclude -origin repos - they're added for .flatpakref files
    # if the remote isn't installed already, and get auto-removed when the app
    # the ref file describes gets uninstalled.  -origin remotes are also marked
    # no-enumerate, so we can filter them here by that.

    existing = False

    url = _get_repofile_repo_url(path)

    if url:
        for remote in fp_sys.list_remotes(None):
            if remote.get_url() == url and not remote.get_noenumerate():
                existing = True
                break

    if existing:
        GObject.idle_add(callback, file, "exists")
        return

    cmd_v = ['flatpak',
             'remote-add',
             '--from',
             path.stem,
             path]

    add_repo_proc = subprocess.Popen(cmd_v, stderr=subprocess.PIPE)
    stdout, stderr = add_repo_proc.communicate()

    if "Error.AccessDenied" in stderr.decode():
        GObject.idle_add(callback, file, "cancel")
        return

    if add_repo_proc.returncode != 0 and "already exists" not in stderr.decode():
        GObject.idle_add(callback, file, "error")
        return

    # We'll do a full cache rebuild - otherwise, after this installer session, the
    # new apps from this remote won't show up until the next scheduled cache rebuild.
    try:
        fp_sys.drop_caches(None)
    except GLib.Error:
        pass

    cache.force_new_cache_async(callback)

def _get_repofile_repo_url(path):
    kf = GLib.KeyFile()

    try:
        if kf.load_from_file(str(path), GLib.KeyFileFlags.NONE):
            url = kf.get_string("Flatpak Repo", "Url")

            return url
    except GLib.Error as e:
        warn(e.message)

    return None



# From flatpak-dir-private.h
"""
/**
 * FLATPAK_DEPLOY_DATA_GVARIANT_FORMAT:
 *
 * s - origin
 * s - commit
 * as - subpaths
 * t - installed size
 * a{sv} - Metadata
 */
"""
FLATPAK_DEPLOY_DATA_GVARIANT_STRING = "(ssasta{sv})"
FLATPAK_DEPLOY_DATA_GVARIANT_FORMAT = GLib.VariantType(FLATPAK_DEPLOY_DATA_GVARIANT_STRING)

def _load_deploy_data(installed_ref):
    deploy_dir = Gio.File.new_for_path(installed_ref.get_deploy_dir())

    data_file = deploy_dir.get_child("deploy")

    try:
        contents, etag = data_file.load_bytes(None)
    except GLib.Error as e:
        warn("Could not load deploy data: %s" % e.message)
        return None

    deploy_data = GLib.Variant.new_from_bytes(FLATPAK_DEPLOY_DATA_GVARIANT_FORMAT, contents, False)

    return deploy_data

def _get_deployed_version(pkginfo):
    iref = get_fp_sys().get_installed_ref(pkginfo.kind,
                                          pkginfo.name,
                                          pkginfo.arch,
                                          pkginfo.branch,
                                          None)

    if not iref:
        return None

    return iref.get_appdata_version()

    # data = _load_deploy_data(iref)

    # metadata = data.get_child_value(4)
    # version_var =  metadata.lookup_value("appdata-version", None)
    # return version_var.get_string()




#!/usr/bin/python3
import threading
import time
import tempfile

import gi
gi.require_version('AppStreamGlib', '1.0')
gi.require_version('Gtk', '3.0')
from gi.repository import GLib, GObject, Gio, Gtk

from . import cache, _flatpak, _apt, dialogs
from .misc import print_timing, check_ml, debug, warn

PKG_TYPE_ALL = None
PKG_TYPE_APT = "a"
PKG_TYPE_FLATPAK = "f"

Gtk.IconTheme.get_default().append_search_path("/usr/share/linuxmint/icons")

class InstallerTask:
    # task types
    INSTALL_TASK = "install"
    UNINSTALL_TASK = "remove"
    UPDATE_TASK = "update"

    # Set after a package selection, reflects whether task can proceed or not
    STATUS_NONE = "none"
    STATUS_OK = "ok"
    STATUS_BROKEN = "broken"
    STATUS_FORBIDDEN = "forbidden"
    STATUS_UNKNOWN = "unknown"

    # Used by standalone progress window to update labels appropriately
    PROGRESS_STATE_INIT = "init"
    PROGRESS_STATE_INSTALLING = "installing"
    PROGRESS_STATE_UPDATING = "updating"
    PROGRESS_STATE_REMOVING = "removing"
    PROGRESS_STATE_FINISHED = "finished"
    PROGRESS_STATE_FAILED = "failed"

    def __init__(self, pkginfo, installer,
                client_info_ready_callback, client_info_error_callback,
                client_installer_finished_cb, client_installer_progress_cb,
                installer_cleanup_cb, installer_error_cleanup_cb, is_addon_task=False, use_mainloop=False, parent_window=None):
        self.type = InstallerTask.INSTALL_TASK

        self.use_mainloop = use_mainloop
        self.parent_window = parent_window

        # pkginfo will be None for an update task
        self.pkginfo = pkginfo
        self.is_addon_task = is_addon_task

        # AsApp if available
        self.asapp = None

        self.name = None

        # Lists the refs we selected for an update, vs those added as dependencies (this is for our confirm dialog to
        # make layout decisions)
        self.initial_refs_to_update = []

        # FlatpakRef : pre-transaction-version or commit
        # For activity logging so we can record pre and post-version.
        self.ref_prior_versions_dict = {}
        self.transaction_log = []

        # Set by .select_pkginfo(), the re-entry point after a task is fully
        # calculated, and the UI should be updated with detailed info about
        # the pending operation (disk use, etc..)
        self.info_ready_callback = client_info_ready_callback
        self.info_error_callback = client_info_error_callback
        # To be checked by the info_ready_callback, to allow the UI to reflect
        # the ability to proceed with a task, or report that something is not right.
        self.info_ready_status = self.STATUS_NONE

        # Set by the backend, the functions to call to actually confirm and perform the
        # task (will be none on STATUS_BROKEN or _FORBIDDEN)
        self.confirm = lambda: True
        self.cancel = lambda: True
        self.execute = None

        # Passed to _flatpak operations to respond to the Cancel button in the
        # standalone progress window. eventually it may be used elsewhere.
        self.cancellable = Gio.Cancellable()

        # Callbacks that will be used at various points during a task being operated on.
        # The .client_* callbacks are arguments of Installer.execute_task().  The
        # client_finished_cb is required.  If the progress callback is missing, a standalone
        # progress window will be provided.
        self.client_progress_cb = client_installer_progress_cb
        self.client_finished_cb = client_installer_finished_cb

        # These are internally used - called as the 'real' error and finished callback,
        # to do some cleanup like removing the task and reloading the apt cache before
        # finally calling task.client_finished_cb
        self.error_cleanup_cb = installer_error_cleanup_cb
        self.finished_cleanup_cb = installer_cleanup_cb

        self.has_window = False
        # Updated throughout a flatpak operation - for now it's used for updating the
        # standalone flatpak progress window
        self.progress_state = self.PROGRESS_STATE_INIT
        # Same - allows the flatpak window to update the current package being installed/removed
        self.current_package_name = None
        # The error message displayed in a popup if a flatpak operation fails.
        self.error_message = None

        self.transaction = None
        self.pkit_request_id = 0

        # The command that can be used to launch the current target package, if it's installed
        self.exec_string = None

        # List of additional packages to install, remove or update, based on the selected
        # pkginfo. Depending on the backend, they will consist of PkPackages or stringified
        # flatpak refs (the result of ref.format_ref()).
        self.to_install = []
        self.to_reinstall = [] # unused
        self.to_remove = []
        self.to_purge = [] # unused
        self.to_update = []
        self.to_downgrade = [] # unused
        self.to_skip_upgrade = [] # unused

        # Size info for display, calculated by the backend during .select_pkginfo()
        self.download_size = 0
        self.install_size = 0
        self.freed_size = 0

        # Static info filled in for display
        if pkginfo:
            self.name = pkginfo.name

            if pkginfo.pkg_hash.startswith("a"):
                self.arch = ""
                self.branch = ""
                self.remote = ""
            else:
                self.arch = pkginfo.arch
                self.remote = pkginfo.remote
                self.branch = pkginfo.branch

    def set_version(self, installer):
        if self.type == InstallerTask.INSTALL_TASK:
            # install packages, show pending version
            self.version = installer.get_version(self.pkginfo)
        else:
            # Remove packages, show current version
            self.version = installer.get_installed_version(self.pkginfo)

    def get_transaction_log(self):
        return self.transaction_log

    def call_info_ready_callback(self):
        if self.info_ready_callback == None:
            return

        if self.use_mainloop:
            GLib.idle_add(self.info_ready_callback, self, priority=GLib.PRIORITY_DEFAULT)
        else:
            self.info_ready_callback(self)

    def handle_error(self, error, info_stage=False):
        try:
            self.error_message = error.message
        except:
            self.error_message = str(error)

        if info_stage:
            if self.info_error_callback == None:
                dialogs.show_error(self.error_message, self.parent_window)
                return

            if self.use_mainloop:
                GLib.idle_add(self.info_error_callback, self, priority=GLib.PRIORITY_DEFAULT)
            else:
                self.info_error_callback(self)
        else:
            dialogs.show_error(self.error_message, self.parent_window)

    def call_finished_cleanup_callback(self):
        if not self.finished_cleanup_cb:
            return

        if self.use_mainloop:
            GLib.idle_add(self.finished_cleanup_cb, self, priority=GLib.PRIORITY_DEFAULT)
        else:
            self.finished_cleanup_cb(self)

    def call_error_cleanup_callback(self):
        if not self.error_cleanup_cb:
            return

        if self.use_mainloop:
            GLib.idle_add(self.error_cleanup_cb, self, priority=GLib.PRIORITY_DEFAULT)
        else:
            self.error_cleanup_cb(self)

class Installer:
    def __init__(self, pkg_type=PKG_TYPE_ALL, temp=False):
        self.tasks = {}
        self.pkg_type = pkg_type

        if temp:
            f = tempfile.NamedTemporaryFile(prefix="mint-common-installer-tmp")
            self.cache_path = f.name
        else:
            self.cache_path = None

        self.remotes_changed = False
        self.inited = False

        self.have_flatpak = False
        self.have_flatpak = self._get_flatpak_status()

        self.cache = {}
        self._init_cb = None

        self.startup_timer = time.time()

    def _get_flatpak_status(self):
        try:
            gi.require_version('Flatpak', '1.0')
            from gi.repository import Flatpak

            return True
        except:
            warn("No flatpak support, install flatpak and gir1.2-flatpak-1.0 and restart mintinstall to enable it.")

        return False

    def init_sync(self):
        """
        Loads the cache synchronously.  Returns True if all went ok, and returns False if there
        is no cache (or it's too old.)  You should then call init() with a callback so the cache
        can be regenerated.
        """
        
        if self.pkg_type == PKG_TYPE_FLATPAK and not self.have_flatpak:
            debug("Not syncing for flatpaks only, as there is currently no support")
            return True

        self.settings = Gio.Settings(schema_id="com.linuxmint.install")

        if self._fp_remotes_have_changed():
            self.remotes_changed = True

        self.backend_table = {}

        self.cache = cache.PkgCache(self.pkg_type, self.cache_path, self.have_flatpak)

        if self.cache.status == self.cache.STATUS_OK and not self.remotes_changed:
            self.inited = True

            GObject.idle_add(self.initialize_appstream)
            self.generate_uncached_pkginfos(self.cache)

            return True

        return False

    def init(self, ready_callback=None):
        """
        Loads the cache asynchronously.  If there is no cache (or it's too old,) it causes
        one to be generated and saved.  The ready_callback is called on idle once this is finished.
        """
        self.backend_table = {}

        self.cache = cache.PkgCache(self.pkg_type, self.cache_path, self.have_flatpak)

        self._init_cb = ready_callback

        if self.cache.status == self.cache.STATUS_OK and not self.remotes_changed:
            self.inited = True

            GObject.idle_add(self._idle_cache_load_done)
        else:
            if self.remotes_changed:
                debug("Installer: Flatpak remotes have changed, forcing a new cache.")

            self.cache.force_new_cache_async(self._idle_cache_load_done)

        return self

    def force_new_cache(self, ready_callback=None):
        """
        Forces the cache to regenerate, calling read_callback when complete
        """
        self.cache.force_new_cache_async(ready_callback)

    def force_new_cache_sync(self):
        """
        Forces the cache to regenerate synchronously
        """
        self.cache.force_new_cache()

    def _idle_cache_load_done(self):
        self.inited = True

        if self.remotes_changed:
            self._store_remotes()
            self.remotes_changed = False

        GObject.idle_add(self.initialize_appstream)

        self.generate_uncached_pkginfos(self.cache)

        debug('Full installer startup took %0.3f ms' % ((time.time() - self.startup_timer) * 1000.0))

        if self._init_cb:
            self._init_cb()

    @print_timing
    def _fp_remotes_have_changed(self):
        """
        We check here for changed remotes.  We care if names, urls, and disabled status changed.
        The 'noenumerate' property won't change, and is usually marked on standalone (-source) ref
        installs.  We don't want to generate a new cache for those - their app can be accessed via
        installed apps, plus if you uninstall the app, the remote gets auto-removed.
        """
        changed = False
        real_remote_count = 0

        saved_remotes = self.settings.get_strv("flatpak-remotes")
        fp_remotes = self.list_flatpak_remotes()

        for remote_info in fp_remotes:
            real_remote_count += 1

            item = "%s::%s::%s" % (remote_info.name, remote_info.url, str(remote_info.disabled))

            if item not in saved_remotes:
                changed = True
                break

        if not changed:
            if len(saved_remotes) != real_remote_count:
                changed = True

        return changed

    @print_timing
    def _store_remotes(self):
        new_remotes = []

        fp_remotes = self.list_flatpak_remotes()

        for remote_info in fp_remotes:
            item = "%s::%s::%s" % (remote_info.name, remote_info.url, str(remote_info.disabled))

            new_remotes.append(item)

        self.settings.set_strv("flatpak-remotes", new_remotes)

    def select_pkginfo(self, pkginfo,
                       client_info_ready_callback, client_info_error_callback,
                       client_installer_finished_cb, client_installer_progress_cb,
                       use_mainloop=False, parent_window=None):
        """
        Initiates calculations for installing or removing a particular package
        (depending upon whether or not the selected package is installed.  Creates
        an InstallerTask instance and populates it with info relevant for display
        and for execution later.  When this is completed, ready_callback is called,
        with the newly-created task as its argument.  Note:  At that point, this is
        the *only* reference to the task object.  It can be safely discarded.  If
        the task is to be run, Installer.execute_task() is called, passing this task
        object, along with callback functions.  The task object is then added to a
        queue (and is tracked in self.tasks from there on out.)
        """
        if pkginfo.pkg_hash in self.tasks.keys():
            task = self.tasks[pkginfo.pkg_hash]

            GObject.idle_add(task.info_ready_callback, task)
            return task.cancellable

        task = InstallerTask(pkginfo, self,
                             client_info_ready_callback, client_info_error_callback,
                             client_installer_finished_cb, client_installer_progress_cb,
                             self._task_finished, self._task_error,
                             use_mainloop=use_mainloop, parent_window=parent_window)

        if self.pkginfo_is_installed(pkginfo):
            # It's not installed, so assume we're installing
            task.type = InstallerTask.UNINSTALL_TASK
        else:
            task.type = InstallerTask.INSTALL_TASK

        task.set_version(self)
        task.asapp = self.get_appstream_app_for_pkginfo(pkginfo)

        if pkginfo.pkg_hash.startswith("a"):
            _apt.select_packages(task)
        else:
            _flatpak.select_packages(task)

        return task.cancellable

    def select_flatpak_updates(self, refs,
                               client_info_ready_callback, client_info_error_callback,
                               client_installer_finished_cb, client_installer_progress_cb,
                               use_mainloop=False):
        """
        Creates an InstallerTask populated with all flatpak packages that can be
        updated. If refs is empty, selects all possible updates, otherwise, only attempts
        to update refs.
        """
        task = InstallerTask(None, self,
                             client_info_ready_callback, client_info_error_callback,
                             client_installer_finished_cb, client_installer_progress_cb,
                             self._task_finished, self._task_error,
                             use_mainloop=use_mainloop)

        task.type = InstallerTask.UPDATE_TASK
        task.initial_refs_to_update = refs if refs else []

        _flatpak.select_updates(task)


    def create_addon_task(self, comp, remote_name, remote_url,
                          client_info_ready_callback, client_info_error_callback,
                          client_installer_finished_cb, client_installer_progress_cb,
                          use_mainloop=False):

        def pkginfo_ready(pkginfo):
            task = InstallerTask(pkginfo, self,
                                 client_info_ready_callback, client_info_error_callback,
                                 client_installer_finished_cb, client_installer_progress_cb,
                                 self._task_finished, self._task_error,
                                 is_addon_task=True,
                                 use_mainloop=use_mainloop)

            if pkginfo.installed:
                task.type = InstallerTask.UNINSTALL_TASK
            else:
                task.type = InstallerTask.INSTALL_TASK

            task.set_version(self)

            _flatpak.select_packages(task)

        def create_pkginfo_thread(comp, remote_name, remote_url):
            pkginfo = _flatpak.create_pkginfo_from_as_component(comp, remote_name, remote_url)
            GLib.idle_add(pkginfo_ready, pkginfo, priority=GLib.PRIORITY_DEFAULT)

        t = threading.Thread(target=create_pkginfo_thread, args=(comp, remote_name, remote_url))
        t.start()

    def list_updated_flatpak_pkginfos(self):
        """
        Returns a list of flatpak pkginfos that can be updated.  Unlike
        prepare_flatpak_update, this is for the convenience of displaying information
        to the user.
        """
        return _flatpak.list_updated_pkginfos(self.cache)

    def find_pkginfo(self, name, pkg_type=PKG_TYPE_ALL, remote=None):
        """
        Attempts to find and return a PkgInfo object, given a package name.  If
        pkg_type is None, looks in first apt, then flatpaks.
        """
        return self.cache.find_pkginfo(name, pkg_type, remote)

    def get_pkginfo_from_ref_file(self, file, ready_callback):
        """
        Accepts a GFile to a .flatpakref on a local path.  If the flatpak's remote
        has not been previously added to the system installation, this also adds
        it and downloads Appstream info as well, before calling ready_callback with
        the created (or existing) PkgInfo as an argument.
        """
        if self.have_flatpak:
            _flatpak.get_pkginfo_from_file(self.cache, file, ready_callback)

    def add_remote_from_repo_file(self, file, ready_callback):
        """
        Accepts a GFile to a .flatpakrepo on a local path.  Adds the remote if it
        doesn't exist already, fetches any appstream data, and then calls
        ready_callback
        """

        if self.have_flatpak:
            _flatpak.add_remote_from_repo_file(self.cache, file, ready_callback)
        else:
            ready_callback(None, "no-flatpak-support")

    def list_flatpak_remotes(self):
        """
        Returns a list of FlatpakRemoteInfos.  The remote_name can be used to match
        with PkgInfo.remote and the title is for display.
        """
        if self.have_flatpak:
            return _flatpak.list_remotes()
        else:
            return []

    def get_remote_info_for_name(self, remote_name):
        if self.have_flatpak:
            for remote in _flatpak.list_remotes():
                if remote.name == remote_name:
                    return remote

        return []

    def pkginfo_is_installed(self, pkginfo):
        """
        Returns whether or not a given package is currently installed.  This uses
        the AptCache or the FlatpakInstallation to check.
        """
        if self.inited:
            if pkginfo.pkg_hash.startswith("a"):
                return _apt.pkginfo_is_installed(pkginfo)
            elif self.have_flatpak and pkginfo.pkg_hash.startswith("f"):
                return _flatpak.pkginfo_is_installed(pkginfo)

        return False

    @print_timing
    def generate_uncached_pkginfos(self, unused=None):
        """
        Flatpaks installed from .flatpakref files may not actually be in the saved
        pkginfo cache, specifically, if they're added from no-enumerate-marked remotes.
        This gets run at startup to collect and generate their info.
        """
        if self.have_flatpak:
            _flatpak.generate_uncached_pkginfos(self.cache)

    @print_timing
    def initialize_appstream(self):
        """
        Loads and caches the AppStream pools so they can be used to provide
        display info for packages.
        """
        if self.have_flatpak:
            _flatpak.initialize_appstream()
        # Is there any reason to use apt's appstream?

    def get_appstream_app_for_pkginfo(self, pkginfo):
        try:
            backend_component = self.backend_table[pkginfo]

            return backend_component
        except KeyError:
            if pkginfo.pkg_hash.startswith("a"):
                backend_component = _apt.search_for_pkginfo_apt_pkg(pkginfo)
            else:
                if self.have_flatpak:
                    backend_component = _flatpak.search_for_pkginfo_as_component(pkginfo)

            if backend_component is not None:
                self.backend_table[pkginfo] = backend_component

            # It's possible at some point we'll refresh appstream at runtime, if so we'll
            # want to clear cached data so it can be re-fetched anew.  For now there's
            # no need. The only possible case is on-demand adding of a remote (from
            # launching a .flatpakref file), and in this case, we won't have had anything
            # cached for it to clear anyhow.

            # pkginfo.clear_cached_info()

            return backend_component

    def get_flatpak_launchables(self, pkginfo):
        """
        Return the launchables associated with the AsApp for this pkginfo.
        """

        if pkginfo.pkg_hash.startswith("a"):
            debug("launch_flatpak: pkginfo is not a flatpak")

        comp = self.get_appstream_app_for_pkginfo(pkginfo)

        if comp is None:
            return None

        launchables = comp.get_launchables()

        if len(launchables) == 0:
            return None

        return launchables

    def get_flatpak_root_path(self):
        """
        Return the root path for the flatpak installation (generally /var/lib/flatpak for system
        and ~/.local/share/flatpak for user.
        """

        return _flatpak.get_fp_sys().get_path().get_path()

    def get_addons(self, pkginfo):
        """
        Returns an array of app ids of names of available addons
        """
        
        if pkginfo.pkg_hash.startswith("a"):
            return None

        comp = self.get_appstream_app_for_pkginfo(pkginfo)

        if comp is None:
            return None

        addons = comp.get_addons()

        if len(addons) == 0:
            return None

        return addons

    def get_display_name(self, pkginfo):
        """
        Returns the name of the package formatted for displaying
        """
        comp = self.get_appstream_app_for_pkginfo(pkginfo)

        return pkginfo.get_display_name(comp)

    def get_summary(self, pkginfo, for_search=False):
        """
        Returns the summary of the package.  If for_search is True,
        this is the raw, unformatted string in the case of apt.
        """
        if for_search and pkginfo.pkg_hash.startswith("a"):
            try:
                return _apt._apt_cache[pkginfo.name].candidate.summary
            except Exception:
                pass

        comp = self.get_appstream_app_for_pkginfo(pkginfo)

        return pkginfo.get_summary(comp)

    def get_description(self, pkginfo, for_search=False):
        """
        Returns the description of the package.  If for_search is True,
        this is the raw, unformatted string in the case of apt.
        """
        if for_search and pkginfo.pkg_hash.startswith("a"):
            try:
                return _apt._apt_cache[pkginfo.name].candidate.description
            except Exception:
                pass

        comp = self.get_appstream_app_for_pkginfo(pkginfo)

        return pkginfo.get_description(comp)

    def get_icon(self, pkginfo, size):
        """
        Returns the icon name (or path) to display for the package
        """
        comp = self.get_appstream_app_for_pkginfo(pkginfo)

        return pkginfo.get_icon(pkginfo, comp, size)

    def get_screenshots(self, pkginfo):
        """
        Returns a list of screenshot urls
        """
        comp = self.get_appstream_app_for_pkginfo(pkginfo)

        return pkginfo.get_screenshots(comp)

    def get_version(self, pkginfo):
        """
        Returns the current version string, if available
        """
        comp = self.get_appstream_app_for_pkginfo(pkginfo)

        return pkginfo.get_version(comp)

    def get_installed_version(self, pkginfo):
        """
        Returns the currently deployed version of a flatpak.
        """
        if pkginfo.pkg_hash.startswith("a"):
            # apt packages we don't really need to make a distinction.
            return self.get_version(pkginfo)
        else:
            # flatpak packages, the appstream component shows the latest version provided in the xml,
            # not the actual installed version.
            return _flatpak._get_deployed_version(pkginfo)

    def get_homepage_url(self, pkginfo):
        """
        Returns the home page url for a package.  If there is
        no url for the package, in the case of flatpak, the remote's url
        is displayed instead
        """
        comp = self.get_appstream_app_for_pkginfo(pkginfo)

        return pkginfo.get_homepage_url(comp)

    def get_help_url(self, pkginfo):
        """
        Returns the help url for a package.  If there is
        no url for the package, returns an empty string. Apt always returns
        an empty string.
        """
        comp = self.get_appstream_app_for_pkginfo(pkginfo)

        return pkginfo.get_help_url(comp)

    def is_busy(self):
        return len(self.tasks.keys()) > 0

    def get_task_count(self):
        return len(self.tasks.keys())

    def get_active_pkginfos(self):
        pkginfos = []

        for pkg_hash in self.tasks.keys():
            pkginfos.append(self.tasks[pkg_hash].pkginfo)

        return pkginfos

    def task_running(self, task):
        """
        Returns whether a given task is currently executing.
        """
        return task.pkginfo.pkg_hash in self.tasks.keys()

    def confirm_task(self, task):
        return task.confirm()

    def cancel_task(self, task):
        task.cancel()

    def execute_task(self, task):
        """
        Executes a given task.  The client_finished_cb is required always, to notify
        when the task completes. The progress and error callbacks are optional.  If
        they're left out, a standalone progress window is created to allow the user to
        see the task's progress (and cancel it if desired.)
        """

        if task.pkginfo != None:
            key = task.pkginfo.pkg_hash
        else:
            key = "updates"

        self.tasks[key] = task

        debug("Starting task for package %s, type '%s'" % (key, task.type))

        task.execute()

    def _task_finished(self, task):
        if not task.pkginfo:
            try:
                del self.tasks["updates"]
                debug("Done with update task (success)")
            except:
                pass
        else:
            key = task.pkginfo.pkg_hash

            if key:
                try:
                    del self.tasks[key]
                    debug("Done with task (success)", key)
                except:
                    pass

        self._post_task_update(task)

    def _task_error(self, task):
        if not task.pkginfo:
            try:
                del self.tasks["updates"]
                debug("Done with update task (failure)")
            except:
                pass
        else:
            key = task.pkginfo.pkg_hash

            if key:
                try:
                    del self.tasks[key]
                    debug("Done with task (failure)", key)
                except:
                    pass

        self._post_task_update(task)

    def _post_task_update(self, task):
        if task.pkginfo and task.pkginfo.pkg_hash.startswith("a"):
            thread = threading.Thread(target=self._apt_post_task_update_thread, args=(task,))
            thread.start()
        else:
            self._run_client_callback(task)

    def _apt_post_task_update_thread(self, task):
        _apt.sync_cache_installed_states()

        # This needs to be called after reloading the apt cache, otherwise our installed
        # apps don't update correctly
        self._run_client_callback(task)

    def _run_client_callback(self, task):
        if task.client_finished_cb:
            GObject.idle_add(task.client_finished_cb, task, priority=GLib.PRIORITY_DEFAULT)



import gi
gi.require_version('XApp', '1.0')
from gi.repository import GLib, Gtk, GObject, Gdk, XApp, Pango

import gettext
APP = 'mint-common'
LOCALE_DIR = "/usr/share/linuxmint/locale"
t = gettext.translation(APP, LOCALE_DIR, fallback=True)
_ = t.gettext

from aptdaemon.gtk3widgets import AptConfirmDialog

######################### Subclass Apt's dialog to keep consistency

class ChangesConfirmDialog(AptConfirmDialog):

    """Dialog to confirm the changes that would be required by a
    transaction.
    """

    def __init__(self, transaction, task=None, parent=None):
        super(ChangesConfirmDialog, self).__init__(transaction, cache=None, parent=parent)
        self.parent_window = parent
        self.set_size_request(500, 350)
        self.task = task

    def _show_changes(self):
        """Show a message and the dependencies in the dialog."""
        self.treestore.clear()
        if not self.parent_window:
            self.set_skip_taskbar_hint(True)
            self.set_keep_above(True)

        # Run parent method for apt
        if not self.task or (self.task.pkginfo and self.task.pkginfo.pkg_hash.startswith("a")):
            """Show a message and the dependencies in the dialog."""
            self.treestore.clear()
            for pkg_list, msg, min_packages in (
                                  [self.task.to_install,      _("Install"), 1 if self.task.type == self.task.INSTALL_TASK else 0],
                                  [self.task.to_reinstall,    _("Reinstall"), 0],
                                  [self.task.to_remove,       _("Remove"), 1 if self.task.type == self.task.UNINSTALL_TASK else 0],
                                  [self.task.to_purge,        _("Purge"), 0],
                                  [self.task.to_update,       _("Upgrade"), 0],
                                  [self.task.to_downgrade,    _("Downgrade"), 0],
                                  [self.task.to_skip_upgrade, _("Skip upgrade"), 0]
                                 ):

                if len(pkg_list) > min_packages:
                    piter = self.treestore.append(None, ["<b>%s</b>" % msg])

                    for pkg in pkg_list:
                        if pkg_list == self.task.to_install and pkg.get_name() == self.task.name:
                            continue

                        self.treestore.append(piter, [pkg.get_name()])
            # If there is only one type of changes (e.g. only installs) expand the
            # tree
            # FIXME: adapt the title and message accordingly
            # FIXME: Should we have different modes? Only show dependencies, only
            #       initial packages or both?
            msg = _("Please take a look at the list of changes below.")
            if len(self.treestore) == 1:
                filtered_store = self.treestore.filter_new(Gtk.TreePath.new_first())
                self.treeview.expand_all()
                self.treeview.set_model(filtered_store)
                self.treeview.set_show_expanders(False)
                if len(self.task.to_install) > 1:
                    title = _("Additional software will be installed")
                elif len(self.task.to_reinstall) > 0:
                    title = _("Additional software will be re-installed")
                elif len(self.task.to_remove) > 0:
                    title = _("Additional software will be removed")
                elif len(self.task.to_purge) > 0:
                    title = _("Additional software will be purged")
                elif len(self.task.to_update) > 0:
                    title = _("Additional software will be upgraded")
                elif len(self.task.to_downgrade) > 0:
                    title = _("Additional software will be downgraded")
                elif len(self.task.to_skip_upgrade) > 0:
                    title = _("Updates will be skipped")
                if len(filtered_store) < 6:
                    self.set_resizable(False)
                    self.scrolled.set_policy(Gtk.PolicyType.AUTOMATIC,
                                             Gtk.PolicyType.NEVER)
                else:
                    self.treeview.set_size_request(350, 200)
            else:
                title = _("Additional changes are required")
                self.treeview.set_size_request(350, 200)
                self.treeview.collapse_all()
            if self.task.download_size > 0:
                msg += "\n"
                msg += (_("%s will be downloaded in total.") %
                        GLib.format_size(self.task.download_size))
            if self.task.freed_size > 0:
                msg += "\n"
                msg += (_("%s of disk space will be freed.") %
                        GLib.format_size(self.task.freed_size))
            elif self.task.install_size > 0:
                msg += "\n"
                msg += (_("%s more disk space will be used.") %
                        GLib.format_size(self.task.install_size))
            self.label.set_markup("<b><big>%s</big></b>\n\n%s" % (title, msg))
        else:
            # flatpak
            self.set_title(_("Flatpaks"))

            min_packages = 1 if self.task.type == self.task.INSTALL_TASK else 0
            if len(self.task.to_install) > min_packages:
                piter = self.treestore.append(None, ["<b>%s</b>" % _("Install")])

                for ref in self.task.to_install:
                    if self.task.pkginfo and self.task.pkginfo.refid == ref.format_ref():
                        continue

                    self.treestore.append(piter, [ref.get_name()])

            min_packages = 1 if self.task.type == self.task.UNINSTALL_TASK else 0
            if len(self.task.to_remove) > min_packages:
                piter = self.treestore.append(None, ["<b>%s</b>" % _("Remove")])

                for ref in self.task.to_remove:
                    if self.task.pkginfo and self.task.pkginfo.refid == ref.format_ref():
                        continue

                    self.treestore.append(piter, [ref.get_name()])

            if len(self.task.to_update) > 0:
                # If this is an update task (like from mintupdate) we may have selected updates explicitly, and there may be
                # updates we *didn't* select but are required for an update we did. We only want to add those updates that
                # are pulled in the second case, since the updates we did select do not need to be displayed again (this is
                # following apt behavior, where we only list dependencies here and unexpected changes).
                header_added = False
                for ref in self.task.to_update:
                    if self.task.type == self.task.UPDATE_TASK:
                        if len(self.task.initial_refs_to_update) == 0 or ref.format_ref() in self.task.initial_refs_to_update:
                            continue

                    if not header_added:
                        piter = self.treestore.append(None, ["<b>%s</b>" % _("Upgrade")])
                        header_added = True

                    self.treestore.append(piter, [ref.get_name()])

            msg = _("Please take a look at the list of changes below.")

            if len(self.treestore) == 1:
                filtered_store = self.treestore.filter_new(
                    Gtk.TreePath.new_first())
                self.treeview.expand_all()
                self.treeview.set_model(filtered_store)
                self.treeview.set_show_expanders(False)

                if len(self.task.to_install) > 1:
                    title = _("Additional software will be installed")
                elif len(self.task.to_remove) > 0:
                    title = _("Additional software will be removed")
                elif len(self.task.to_update) > 0:
                    title = _("Additional software will be upgraded")

                if len(filtered_store) < 6:
                    self.set_resizable(False)
                    self.scrolled.set_policy(Gtk.PolicyType.AUTOMATIC,
                                             Gtk.PolicyType.NEVER)
                else:
                    self.treeview.set_size_request(350, 200)
            else:
                title = _("Additional changes are required")
                self.treeview.set_size_request(350, 200)
                self.treeview.collapse_all()

            if self.task.download_size > 0:
                msg += "\n"
                msg += (_("%s will be downloaded in total.") %
                        GLib.format_size(self.task.download_size))
            if self.task.freed_size > 0:
                msg += "\n"
                msg += (_("%s of disk space will be freed.") %
                        GLib.format_size(self.task.freed_size))
            elif self.task.install_size > 0:
                msg += "\n"
                msg += (_("%s more disk space will be used.") %
                        GLib.format_size(self.task.install_size))
            self.label.set_markup("<b><big>%s</big></b>\n\n%s" % (title, msg))

    def map_package(self, pkg):
        """Map a package to a different object type, e.g. applications
        and return a list of those.

        By default return the package itself inside a list.

        Override this method if you don't want to store package names
        in the treeview.
        """
        return [pkg]

    def render_package_desc(self, column, cell, model, iter, data):
        value = model.get_value(iter, 0)

        cell.set_property("markup", value)


class FlatpakProgressWindow(Gtk.Dialog):
    """
    Progress dialog for standalone flatpak installs, removals, updates.
    Intended to be used when not working as part of a parent app (like mintinstall)
    """

    def __init__(self, task, parent=None):
        Gtk.Dialog.__init__(self, parent=parent)
        self.set_default_size(400, 140)
        self.task = task
        self.finished = False

        # Progress goes directly to this window
        task.client_progress_cb = self.window_client_progress_cb

        # finished callbacks route thru the installer
        # but we want to see them in this window also.
        self.final_finished_cb = task.client_finished_cb
        task.client_finished_cb = self.window_client_finished_cb
        self.pulse_timer = 0

        self.real_progress_text = None

        # Setup the dialog
        self.set_border_width(6)
        self.set_resizable(False)
        self.get_content_area().set_spacing(6)
        # Setup the cancel button
        self.button = Gtk.Button.new_from_stock(Gtk.STOCK_CANCEL)
        self.button.set_use_stock(True)
        self.get_action_area().pack_start(self.button, False, False, 0)
        self.button.connect("clicked", self.on_button_clicked)
        self.button.show()

        # labels and progressbar
        vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        vbox.set_spacing(12)
        vbox.set_border_width(10)

        self.label = Gtk.Label(max_width_chars=45)
        vbox.pack_start(self.label, False, False, 0)
        self.label.set_halign(Gtk.Align.START)
        self.label.set_line_wrap(True)

        self.progress = Gtk.ProgressBar()
        vbox.pack_end(self.progress, False, True, 0)
        self.get_content_area().pack_start(vbox, True, True, 0)

        self.set_title(_("Flatpak"))
        XApp.set_window_icon_name(self, "system-software-installer")

        vbox.show_all()
        self.realize()

        self.progress.set_size_request(350, -1)
        functions = Gdk.WMFunction.MOVE | Gdk.WMFunction.RESIZE
        try:
            self.get_window().set_functions(functions)
        except TypeError:
            # workaround for older and broken GTK typelibs
            self.get_window().set_functions(Gdk.WMFunction(functions))

        # catch ESC and behave as if cancel was clicked
        self.connect("delete-event", self._on_dialog_delete_event)

    def start_progress_pulse(self):
        if self.pulse_timer > 0:
            return

        self.progress.pulse()
        self.pulse_timer = GObject.timeout_add(1050, self.progress_pulse_tick)

    def progress_pulse_tick(self):
        self.progress.pulse()

        return GLib.SOURCE_CONTINUE

    def stop_progress_pulse(self):
        if self.pulse_timer > 0:
            GObject.source_remove(self.pulse_timer)
            self.pulse_timer = 0

    def _on_dialog_delete_event(self, dialog, event):
        self.button.clicked()
        return True

    def window_client_progress_cb(self, pkginfo, progress, estimating, status_text):
        if estimating:
            self.start_progress_pulse()
        else:
            self.stop_progress_pulse()

            self.progress.set_fraction(progress / 100.0)
            XApp.set_window_progress(self, progress)

        self.label.set_text(status_text)

    def window_client_finished_cb(self, task):
        self.finished = True

        self.destroy()
        self.final_finished_cb(task)

    def on_button_clicked(self, button):
        if not self.finished:
            self.task.cancel()

def show_error(message, parent_window=None):
    Gdk.threads_add_idle(GLib.PRIORITY_DEFAULT_IDLE, _show_error_mainloop, message, parent_window)

def _show_error_mainloop(message, parent_window):
    dialog = Gtk.MessageDialog(None,
                               Gtk.DialogFlags.DESTROY_WITH_PARENT,
                               Gtk.MessageType.ERROR,
                               Gtk.ButtonsType.OK,
                               "")
    if parent_window is not None:
        dialog.set_transient_for(parent_window)
        dialog.set_modal(True)
    dialog.set_title(GLib.get_application_name())

    text = _("An error occurred")
    dialog.set_markup("<big><b>%s</b></big>" % text)

    scroller = Gtk.ScrolledWindow(min_content_height = 75, max_content_height=400, min_content_width=400, propagate_natural_height=True)
    dialog.get_message_area().pack_start(scroller, False, False, 8)

    message_label = Gtk.Label(message, lines=20, wrap=True, wrap_mode=Pango.WrapMode.WORD_CHAR, selectable=True)
    message_label.set_max_width_chars(60)
    message_label.show()
    scroller.add(message_label)

    dialog.show_all()
    dialog.run()
    dialog.destroy()

    return GLib.SOURCE_REMOVE




import time
import threading
import apt

import gi
gi.require_version('AppStream', '1.0')
gi.require_version('Gtk', '3.0')
gi.require_version('Gdk', '3.0')
gi.require_version("PackageKitGlib", "1.0")

from gi.repository import Gtk, GLib
from gi.repository import PackageKitGlib as packagekit

from aptdaemon.gtk3widgets import AptProgressDialog

from .pkgInfo import AptPkgInfo
from .dialogs import ChangesConfirmDialog
from .misc import check_ml, warn, debug
from . import dialogs

# List extra packages that aren't necessarily marked in their control files, but
# we know better...
CRITICAL_PACKAGES = ["mint-common", "mint-meta-core", "mintdesktop", "python3"]

def capitalize(string):
    if len(string) > 1:
        return (string[0].upper() + string[1:])

    return (string)

_apt_cache = None
_apt_cache_lock = threading.Lock()
_as_pool = None

def get_apt_cache(full=False):
    global _apt_cache

    if full or (not _apt_cache):
        with _apt_cache_lock:
            _apt_cache = apt.Cache()

    return _apt_cache

def add_prefix(name):
    return "apt:%s" % (name)

def get_real_error(code):
    n = int(code)
    if n > 255:
        return packagekit.ErrorEnum(n - 255)
    else:
        return packagekit.ClientError(n)

def make_pkg_hash(apt_pkg):
    if not isinstance(apt_pkg, apt.Package):
        raise TypeError("apt.make_pkg_hash_make must receive apt.Package, not %s" % type(apt_pkg))

    return add_prefix(apt_pkg.name)

def process_full_apt_cache(cache):
    apt_time = time.time()
    apt_cache = get_apt_cache()

    sections = {}

    keys = apt_cache.keys()

    for key in keys:
        name = apt_cache[key].name
        pkg = apt_cache[key]

        if name.startswith("lib") and not name.startswith(("libreoffice", "librecad", "libk3b7", "libimage-exiftool-perl")):
            continue
        if name.endswith(":i386") and name != "steam:i386":
            continue
        if name.endswith("-dev"):
            continue
        if name.endswith("-dbg"):
            continue
        if name.endswith("-doc"):
            continue
        if name.endswith("-common"):
            continue
        if name.endswith("-data"):
            continue
        if "-locale-" in name:
            continue
        if "-l10n-" in name:
            continue
        if name.endswith("-dbgsym"):
            continue
        if name.endswith("l10n"):
            continue
        if name.endswith("-perl"):
            continue
        if name == "snapd":
            continue
        if name == "pepperflashplugin-nonfree": # formerly marked broken, it's now a dummy and has no dependents (and only exists in Mint 20).
            continue
        if ":" in name and name.split(":")[0] in keys:
            continue
        try:
            if "transitional" in pkg.candidate.summary.lower():
                continue
        except Exception as e:
            warn("Problem parsing package (maybe it's virtual): %s: %s" % (name, e))
            continue
            # pass

        pkg_hash = make_pkg_hash(pkg)

        if pkg.candidate:
            section_string = pkg.candidate.section

            if "/" in section_string:
                section = section_string.split("/")[1]
            else:
                section = section_string

        try:
            sections[section].append(pkg_hash)
        except Exception:
            sections[section] = []
            sections[section].append(pkg_hash)

        cache[pkg_hash] = AptPkgInfo(pkg_hash, pkg)

    debug('Installer: Processing APT packages for cache took %0.3f ms' % ((time.time() - apt_time) * 1000.0))

    return cache, sections

# def initialize_appstream():
#     global _as_pool

#     if _as_pool == None:
#         pool = AppStream.Pool()
#         pool.set_cache_flags(AppStream.CacheFlags.NONE)
#         pool.load()
#         _as_pool = pool

def search_for_pkginfo_apt_pkg(pkginfo):
    name = pkginfo.name

    apt_cache = get_apt_cache()

    try:
        return apt_cache[name]
    except:
        return None

def find_pkginfo(cache, string):
    try:
        pkginfo = cache[add_prefix(string)]
    except:
        pkginfo = None

    return pkginfo

def pkginfo_is_installed(pkginfo):
    global _apt_cache_lock
    apt_cache = get_apt_cache()

    with _apt_cache_lock:
        try:
            return apt_cache[pkginfo.name].installed != None
        except:
            return False

def sync_cache_installed_states():
    get_apt_cache(full=True)

def select_packages(task):
    task.transaction = MetaTransaction(task)

    debug("Installer: Calculating changes required for APT package: %s" % task.pkginfo.name)

class MetaTransaction(packagekit.Task):
    def __init__(self, task):
        packagekit.Task.__init__(self)

        self.task = task
        self.simulated_download_size = 0

        thread = threading.Thread(target=self._calculate_apt_changes)
        thread.start()

    def _calculate_apt_changes(self):
        global _apt_cache_lock
        apt_cache = get_apt_cache()
        with _apt_cache_lock:
            apt_cache.clear()
            apt_pkg = apt_cache[self.task.pkginfo.name]
            results = None

            pkg_id = packagekit.Package.id_build(apt_pkg.shortname, "", apt_pkg.architecture(), "")

        self.set_simulate(True)

        try:
            if self.task.type == "remove":
                results = self.remove_packages_sync(
                    [pkg_id],
                    True, True, # allow_deps, autoremove
                    self.task.cancellable,  # cancellable
                    self.on_transaction_progress,
                    None  # progress data
                )
            elif self.task.type == "install":
                results = self.install_packages_sync(
                    [pkg_id],
                    self.task.cancellable,  # cancellable
                    self.on_transaction_progress,
                    None  # progress data
                )
            elif self.task.type == "update":
                debug("todo update")
        except GLib.Error as e:
            self.on_transaction_error(e)

        self.on_transaction_finished(results)

    def on_transaction_error(self, error):
        # PkErrorEnums are sent from the backend mainly
        # PkClientErrors are related to interaction with a task/client - accept/deny, etc...
        if error.code == packagekit.ClientError.DECLINED_SIMULATION:
            # canceled via additional-changes dialog
            return

        # it thinks it's a PkClientError but it's really PkErrorEnum
        # the GError code is set to 0xFF + code
        real_code = error.code
        if error.code >= 0xFF:
            real_code = error.code - 0xFF

            if real_code == packagekit.ErrorEnum.NOT_AUTHORIZED:
                # Silently ignore auth failures or cancellation.
                return

        if self.task.cancellable.is_cancelled():
            # user navigated away before simulation was complete, etc...
            return

        if real_code == packagekit.ErrorEnum.CANNOT_REMOVE_SYSTEM_PACKAGE or self.task.pkginfo.name in CRITICAL_PACKAGES:
            self.task.info_ready_status = self.task.STATUS_FORBIDDEN

        if self.task.info_ready_status == self.task.STATUS_NONE:
            self.task.info_ready_status = self.task.STATUS_UNKNOWN

        self.task.handle_error(error, info_stage = self.get_simulate())

    def on_transaction_finished(self, results):
        # == operation was successful
        if results:
            exit_code = results.get_exit_code()
            pkerror = results.get_error_code()
            if pkerror:
                warn("Finished code: ", pkerror.get_code(), pkerror.get_details())
            debug("Exit code:", exit_code)

        if self.task.error_message:
            self.task.call_error_cleanup_callback()
        else:
            self.task.call_finished_cleanup_callback()

    def on_transaction_progress(self, progress, ptype, data=None):
        if progress.get_status() == packagekit.StatusEnum.UNKNOWN:
            return

        if self.get_simulate():
            if ptype == packagekit.ProgressType.DOWNLOAD_SIZE_REMAINING:
                new_size = progress.get_download_size_remaining()
                if new_size > self.simulated_download_size:
                    self.simulated_download_size = new_size
                # print("current:", progress.get_package_id(), progress.get_status())
            return

        if ptype == packagekit.ProgressType.PERCENTAGE:
            if self.task.client_progress_cb:
                GLib.idle_add(self.task.client_progress_cb,
                              self.task.pkginfo,
                              progress.get_percentage(),
                              False,
                              priority=GLib.PRIORITY_DEFAULT)

    def do_simulate_question(self, request, results):
        if self.task.cancellable.is_cancelled():
            self.user_declined()
            return;

        self.task.pkit_request_id = request
        sack = results.get_package_sack()

        install_dbginfo = []
        remove_dbginfo = []
        update_dbginfo = []
        added_size = 0
        freed_size = 0

        global _apt_cache_lock
        apt_cache = get_apt_cache()

        with _apt_cache_lock:
            for pkg in sack.get_array():
                info = pkg.get_info()

                def calc_space(pkg, is_update=False):
                    apt_pkg = apt_cache["%s:%s" % (pkg.get_name(), pkg.get_arch())]

                    candidate = apt_pkg.candidate

                    if is_update:
                        for version in apt_pkg.versions:
                            if version.is_installed:
                                return candidate.installed_size - version.installed_size

                    return candidate.installed_size

                if info == packagekit.InfoEnum.INSTALLING:
                    self.task.to_install.append(pkg)
                    added_size += calc_space(pkg)
                    install_dbginfo.append("%s:%s (%s)" % (pkg.get_name(), pkg.get_arch(), pkg.get_version()))
                elif info == packagekit.InfoEnum.UPDATING:
                    self.task.to_update.append(pkg)
                    added_size += calc_space(pkg, is_update=True)
                    update_dbginfo.append("%s:%s (%s)" % (pkg.get_name(), pkg.get_arch(), pkg.get_version()))
                elif info == packagekit.InfoEnum.REMOVING:
                    self.task.to_remove.append(pkg)
                    freed_size += calc_space(pkg)
                    remove_dbginfo.append("%s:%s (%s)" % (pkg.get_name(), pkg.get_arch(), pkg.get_version()))

            debug("For install:", install_dbginfo)
            debug("For removal:", remove_dbginfo)
            debug("For upgrade:", update_dbginfo)

            self.task.download_size = self.simulated_download_size

            space = added_size - freed_size

            if space < 0:
                self.task.freed_size = space * -1
                self.task.install_size = 0
            else:
                self.task.freed_size = 0
                self.task.install_size = space

            for pkg in self.task.to_remove:
                apt_pkg_name = apt_cache["%s:%s" % (pkg.get_name(), pkg.get_arch())]

                if self._is_critical_package(apt_cache[apt_pkg_name]):
                    warn("Installer: apt - cannot remove critical package: %s" % apt_pkg_name)
                    self.task.info_ready_status = self.task.STATUS_FORBIDDEN

            if self.task.info_ready_status not in (self.task.STATUS_FORBIDDEN, self.task.STATUS_BROKEN):
                self.task.info_ready_status = self.task.STATUS_OK
                self.task.confirm = self._confirm_transaction
                self.task.cancel = self._cancel_transaction
                self.task.execute = self._execute_transaction

            self.task.call_info_ready_callback()

    def _is_critical_package(self, pkg):
        try:
            if pkg.versions[0].priority == "required" or pkg.name in CRITICAL_PACKAGES:
                return True

            return False
        except Exception:
            return False

    def _confirm_transaction(self):
        if len(self.task.to_install) > 1 or len(self.task.to_remove) > 1 or len(self.task.to_update) > 0:
            dia = ChangesConfirmDialog(self, self.task, parent=self.task.parent_window)
            res = dia.run()
            dia.hide()
            dia.destroy()

            return res == Gtk.ResponseType.OK
        else:
            return True

    def _cancel_transaction(self):
        self.task.cancellable.cancel()

        if self.task.pkit_request_id > 0:
            self.user_declined(self.task.pkit_request_id)
            self.task.pkit_request_id = 0

    def _execute_transaction(self):
        self.set_simulate(False)

        if self.task.cancellable.is_cancelled():
            return

        if self.task.client_progress_cb != None:
            self.task.has_window = True

        if self.task.has_window:
            self.user_accepted(self.task.pkit_request_id)
        else:
            progress_window = AptProgressDialog(self)
            progress_window.run(show_error=False, error_handler=self._on_error)
